'use strict';

const exec = require('child_process').exec,
      _ = require('lodash'),
      fs = require('fs');

const allowedTypes = [
    'feat',     // (new feature for the user, not a new feature for build script)
    'fix',      // (bug fix for the user, not a fix to a build script)
    'docs',     // (changes to the documentation)
    'style',    // (formatting, missing semi colons, etc; no production code change)
    'refactor', // (refactoring production code, eg. renaming a variable)
    'test',     // (adding missing tests, refactoring tests; no production code change)
    'chore'     // (updating grunt tasks etc; no production code change)
];

const irrelevantTypes = [
    'docs', 'style', 'test', 'chore'
];

const markdownTypes = {
    'feat':     'New Features',
    'fix':      'Bug Fixes',
    'docs':     'Documentation',
    'style':    'Code Formatting',
    'refactor': 'Code Refactor',
    'test':     'Test Changes',
    'chore':    'Other Changes'
};

const commitFields = {
    shortHash: '%h',
    subject: '%s',
    commitDate: '%cd'
};

const COMMIT_END = '::end::',
      FIELD_DELIMITER = '::del::';
      
const defaultOpts = {
    fields: ['shortHash', 'subject', 'commitDate'],
    ignoreIrrelevants: true,
    sort: true
};

const generateCommandGrep = function(command, options) {
    var types = _.cloneDeep(allowedTypes);
    
    if (options.ignoreIrrelevants) {
        _.remove(types, item => {
            if (irrelevantTypes.indexOf(item) !== -1) {
                return true;
            } else {
                return false;
            }
        });
    }
    
    command += ' --grep "';
    
    for (let i = 0; i < types.length; i++) {
        if (i !== 0) command += '\\|';
        
        command += `${types[i]}(\\|${types[i]}:`;
    }
    
    command += '"';
    
    return command;
};

const generateCommandPretty = function(command, options) {
    var fields = options.fields;
    
    command += ' --pretty="';
    
    for (let i = 0; i < fields.length; i++) {
        // %h${FIELD_DELIMITER}%s${FIELD_DELIMITER}%cd${COMMIT_END}
        
        if (i !== 0) command += FIELD_DELIMITER;
        
        command += commitFields[fields[i]];
    }
    
    command += `${COMMIT_END}"`;
    
    return command;
};

const executeGitCommand = function(command, options, sync, callback) {
    if (!sync) {
        exec(command, (err, stdout, stderr) => {
            if (err || stderr) {
                callback(stderr || err, {
                    done: false,
                    data: null
                });
            }
            
            let commits = parseStdout(stdout, options.fields);
            let result = _.cloneDeep(commits);
            
            if (options.markdown || !!options.file) {
                result = generateMarkdown(commits, options);
                
                if (!!options.file) {
                    return saveToFile(result, options.file, sync, callback);
                }
            }
            
            callback(null, {
                done: true,
                data: result
            });
        });
    } else {
        return 'git log --grep feat';
        //result = saveToFile(result, options.file, sync);
    }
};

const parseStdout = function(stdout, fields) {
    var str = stdout.split(COMMIT_END);
    
    var commits = {};
    str.forEach(commit => {
        if (commit && commit.trim().length) {
            let parsed = parseCommit(commit.trim(), fields);
            
            if (!commits[parsed.type]) commits[parsed.type] = [];
            
            commits[parsed.type].push(parsed);
        }
    });
    
    return commits;
};

const parseCommit = function(raw, fields) {
    //':start::del:8bd5625:del:feat(app): Create the application project structure:del:Tue Oct 4 12:29:00 2016 +0000:del::end:'
    var commit = {};
    
    var data = raw.split(FIELD_DELIMITER);
    
    for (let i = 0; i < fields.length; i++) {
        commit[fields[i]] = data[i];
    }
    
    let type = commit.subject.split(':')[0].trim(),
        msg = commit.subject.split(':')[1].trim(),
        scope = type.match(/\(([\s\S]*)\)$/);
    
    commit.type = type.split('(')[0];
    
    if (scope !== null) commit.scope = scope[1];
    
    commit.subjectMessage = msg;
    
    return commit;
};

const generateMarkdown = function(data, options) {
    var markdown = '';
    
    var types = Object.keys(data);
    
    if (options.sort) {
        types = [];
        
        allowedTypes.forEach(type => {
            if (!!data[type]) types.push(type);
        });
    }
    
    //for (let type in data) {
    types.forEach(type => {
        let commits = data[type];
        
        markdown += `#### ${markdownTypes[type]}\n`;
        
        commits.forEach(commit => {
            if (commit) {
                let line = '* ';
                
                if (commit.scope) {
                    line += `**${commit.scope}**: `;
                }
                
                // Adding the commit message
                line += commit.subjectMessage;
                
                // Adding the commit link
                if (options.remote) {
                    let link = options.remote;
                    
                    if (link.slice(-1) === '/') link = link.slice(0, -1);
                    
                    if (options.remote_github) link += '/commit';
                    if (options.remote_bitbucket) link += '/commits';
                    
                    line += ` ([${commit.shortHash}](${link}/${commit.shortHash}))`;
                }
                
                // Adding closed issues
                
                markdown += `${line}\n`;
            }
        });
        
        markdown += '\n';
    // }
    });
    
    markdown = markdown.trim();
    
    // Adding separation
    markdown += '\n-----\n';
    
    var today = new Date();
    markdown += `*Changelog generated at ${today.toUTCString()} by [GitCommitMarkdown](https://github.com/EastolfiWebDev/GitCommitMarkdown)*`;
    
    return markdown;
};

const saveToFile = function(data, file, sync, callback) {
    if (_.isBoolean(file)) file = 'CHANGELOG.md';
    
    if (!sync) {
        fs.writeFile(file, data, err => {
            if (err) {
                callback(err, {
                    done: false,
                    error: err
                });
            } else {
                callback(null, {
                    done: true,
                    data: `File "${file}" created`
                });
            }
        });
    } else {
        try {
            fs.writeFileSync(file, data);
            
            return {
                done: true,
                data: `File "${file}" created`
            };
        } catch (err) {
            return {
                done: false,
                error: err
            };
        }
    }
};

function gitCommitMarkdown(options, callback) {
    var sync = false;
    var command = 'git log';
    
    _.assign(options, defaultOpts);
    
    if (!callback) sync = true;
    
    // Only accept a string as remote repository (github and bitbucket currently)
    if (options.remote) {
        if (_.isString(options.remote) && options.remote.match(/github|bitbucket/i)) {
            options.remote_github = !!options.remote.match(/github/i);
            options.remote_bitbucket = !!options.remote.match(/bitbucket/i);
        } else {
            throw new Error('Remote repository only supported as "https://github.com/..." or "https://bitbucket.org/..."');
        }
    }
    
    command = generateCommandGrep(command, options);
    
    command = generateCommandPretty(command, options);

    return executeGitCommand(command, options, sync, callback);
}

module.exports = gitCommitMarkdown;